/* This tool was developed at the University of Delaware and is publicly
   available for academic and non-commercial use. University of Delaware
   assumes no responsibility for consequences arising from the use of
   this tool. It is released under the GPL license. All questions and
   feedback should be directed to Jelena Mirkovic <sunshine@cis.udel.edu> */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "customize.h"
#include "hash.h"
#include "constants.h"

#include "dest.h"
#include "conn.h"

#include "pcap_packet_read.h"

FILE* victimfile;
int checkprivateIPflag = TRUE;
double confidence[]={0, 1, 2, 3, 5, 5, 3, 3, 3.5, 4, 5};

double resetTime = 0, cleanTime = 0;

int counter = 0;
int first = TRUE;
extern int visits, rounds, inserts, finds;

struct hash_table* conntable, *desttable;

/* Find if there is a tcp connection between these two addresses 
   input: source IP address, destination IP address
   output: returns TRUE if there's a TCP connection between source 
   and destination, o/w returns FALSE */

int findTCP(unsigned int source, unsigned int dest)
{
  int k;

  /* For each valid entry in the connection table */
  for(k = 0; k < conntable->size; k++)
    {
      if (conntable->entries[k].state == VALID)
	{
	  struct conn_key* ckey = (struct conn_key*)
	    conntable->entries[k].key;
	  struct conn_data* cdata = (struct conn_data*)
	    conntable->entries[k].data;
	  
	  if ((ckey->ip1 == source && ckey->ip2 == dest)||
	      (ckey->ip2 == source && ckey->ip1 == dest))
	    if (cdata->protocol == TCP)
	      return TRUE;
	}
    }
  return FALSE;
}


/* Calculates the average traffic in the window
   input: pointer to integer array, low index and high index for the window
   output: average in the window */

double rateinwindow(int* array, int from, int to)
{
  int i;
  int sum = 0;

  for(i=from; i<=to;i++)
    sum += array[i];
  return (double)sum/(to-from+1);
}

/* Tests if IP address is private address
   input: IP address in decimal form
   output: returns TRUE if IP address is private, o/w returns FALSE */


int privateIP(unsigned int ip)
{
  if ((ip >> 24) == 10)
    return TRUE;
  else if((ip >> 24) == 172)
    {
      int second = ((ip << 8) >> 24);
      if((second >= 16) && (second <= 31))
	return TRUE;
      else 
	return FALSE;
    }
  else if((ip >> 24) == 192)
    {
      int second = ((ip << 8) >> 24);
      if(second == 168)
	return TRUE;
      else 
	return FALSE;
    }
  else
    return FALSE;
}

/* Resets variables in destination and connection table entries, deletes idle 
   connections from connection table 
   input: Pointer to destination table, pointer to connection table, time
   output: Connection table and destination table will be updated */

void resetCounters(struct hash_table* desttable,struct hash_table* conntable, 
		   double tracetime)
{
  int k, i;

  printf("Resetting\n");

  /* For each valid entry in the destination table */
  for(k = 0; k < desttable->size; k++)
    {
      if (desttable->entries[k].state == VALID)
	{
	  int curUDP;
	  struct dest_key* dkey = (struct dest_key*)
	    desttable->entries[k].key;
	  struct dest_data* ddata = (struct dest_data*)
	    desttable->entries[k].data;

	  for(i=0;i<NUM_PROTOCOLS;i++)
	    {
	      int j;

	      /* Initialize traffic variables for the destination */
	      ddata->packetssent[i] *= BETA;
	      ddata->bytessent[i] *= BETA;
	      ddata->packetsreceived[i] *= BETA;
	      ddata->bytesreceived[i] *= BETA;

	      /* Shift the rate array */
	      for(j=2*WSIZE-1;j>0;j--)
		ddata->bytesreceivedarray[i][j] = ddata->bytesreceivedarray[i][j-1]; 
	      ddata->bytesreceivedarray[i][0] = 0;
	    }

	  /* Reset attack flags */
	  for(i=0;i<NO_OF_ATTACKS;i++)
	    if (i != UDP_FLOOD)
	      ddata->victimtype[i] = FALSE;

	  /* Check if we can reset UDP flag */
	  curUDP = rateinwindow(ddata->bytesreceivedarray[UDPI],0, WSIZE);
	  if (curUDP < ddata->stoprate)
	    {
	      ddata->victimtype[UDP_FLOOD] = FALSE;
	    }
	}
    }

  /* For each valid entry in the connection table */
  for(k = 0; k < conntable->size; k++)
    {
      if (conntable->entries[k].state == VALID)
	{
	  struct conn_key* ckey = (struct conn_key*)
	    conntable->entries[k].key;
	  struct conn_data* cdata = (struct conn_data*)
	    conntable->entries[k].data;

	  cdata->oop = 0;
	  cdata->retransmissions = 0;

	  /* Delete connection if it is idle for a period */
	  if (tracetime - cdata->lasttimestamp > CONN_STALE)
	    {
	      delete(ckey, conntable);
	    }
	}
    }
}


/* Initializes all the attributes in the destination table entry 
   input: pointer to destination table entry and current time
   output: initalized values of destination table entry */

void init_dest(struct dest_data * data,double time)
{
  int i;

  for(i=0;i<NUM_PROTOCOLS;i++)
    {
      int j;

      data->packetssent[i] = 0;
      data->bytessent[i] = 0;
      data->packetsreceived[i] = 0;
      data->bytesreceived[i] = 0;
      for(j=0;j<2*WSIZE;j++)
	data->bytesreceivedarray[i][j] = 0;
    }
  data->tcprto = 0;
  data->icmprto = 0;
  data->synrto = 0;
  data->stoprate = 0;
  data->synsent = 0;
  data->synacksent = 0;
  data->synreceived = 0;
  data->synackreceived = 0;
  data->lasttimestamp = time;
}

/* Initializes all the attributes in the connection table entry 
   input: pointer to connection table entry, current time, direction of the connection(reverse or not) 
   output: initalized values of connection table entry */


void init_conn(struct conn_data * data)
{
  data->legitimate = TRUE;
  data->lastseqnum = INIT_SEQ_NUM;
  data->firstseqnum = INIT_SEQ_NUM;
  data->retransmissions = 0;
  data->nooffragmented = 0;
  data->handshake = 0;
  data->oneway = TRUE;
  data->oop = 0;
}

/* Legitimacy of connection is checked, sequence numbers, suspicious points, 
   number of retransmissions for the connection are updated.
   input: pointer to the entry in the connection table, sequence number of
   the packet, length of the packet
   output: sequence numbers,suspicious points, number of retransmissions
   for the connection are updated, legitimacy of the connection is checked */

void check_legitimate(struct conn_data* cdata, unsigned int seqnum, int len)
{
  if (cdata->legitimate == FALSE)
    return;
  
  /* Update seqnum's for the connection */
  if(seqnum == cdata->lastseqnum || cdata->firstseqnum == INIT_SEQ_NUM)
    {
      cdata->firstseqnum = seqnum;
      cdata->lastseqnum = seqnum + len;
    }
  else
    /* Check for same packets */
    if (seqnum == cdata->firstseqnum && seqnum+len == cdata->lastseqnum)
      {
	cdata->retransmissions ++;
      }

  /* Packets that are not same but out of order */
    else
      {
	cdata->oop += (seqnum - cdata->lastseqnum);
	if (seqnum > cdata->lastseqnum)
	  {
	    cdata->firstseqnum = seqnum;
	    cdata->lastseqnum = seqnum+len;
	  }
      }

  /* If suspicious points exceed threshold */ 
  if(abs(cdata->oop) > MAX_SEQ_DIFF)
    cdata->legitimate = FALSE;

  /* If retransmissions exceed threshold */
  if (cdata->retransmissions > MAX_RETRANSMISSIONS)
    cdata->legitimate = FALSE;
}

/* Progressively removes connections that sent little bit of data
   until it frees enough space in connection table
   input: index of the original connection(needs to be checked to prevent 
   it from getting deleted)
   output: connections that sent small amount of data are deleted */

void emergencyCleanUpConnTable(int location)
{
  int k;
  int sent = 5;

  printf("Performing emergency cleanup\n");

  /* Progressively removes connections that send little bit
     of data until enough space is acquired */
  while(conntable->filled >= 0.7*conntable->size)
    {
      sent = sent*2;
      
      /* For each valid entry in the connection table */
      for(k = 0; k < conntable->size; k++)
	{
	  if (conntable->entries[k].state == VALID)
	    {
	      struct conn_key* ckey = (struct conn_key*)
		conntable->entries[k].key;
	      struct conn_data* cdata = (struct conn_data*)
		conntable->entries[k].data;
	      
	      /* Check if it sent less than the threshold or original connection */
	      if (cdata->sent < sent && k!= location)
		{
		  delete(ckey, conntable);
		}
	    }
	}
    }
}

/* Progressively removes destinations that sent little bit of data
   until it frees enough space in destination table
   input: index of the original destination(needs to be checked to prevent 
   it from getting deleted)
   output: destinations that sent small amount of data are deleted */

void emergencyCleanUpDestTable(int location)
{
  int k;
  int received = 5;

  printf("Performing emergency cleanup\n");

  /* Progressively removes destinations that received little bit
     of data until enough space is acquired */
  while(desttable->filled >= 0.7*desttable->size)
    {
      received = received*2;
  
      /* For each valid entry in the connection table */
      for(k = 0; k < desttable->size; k++)
	{
	  if (desttable->entries[k].state == VALID)
	    {
	      struct dest_key* dkey = (struct dest_key*)
		desttable->entries[k].key;
	      struct dest_data* ddata = (struct dest_data*)
		desttable->entries[k].data;
	      int i, sum = 0;

	      /* Sum of the bytes of data that destination received is calculated */
	      for(i=0;i<NUM_PROTOCOLS;i++)
		sum += ddata->bytesreceived[i];

	      /* Check if destination received less than the threshold or original connection */
	      if (sum < received && k!= location)
		{
		  delete(dkey, desttable);
		}
	    }
	}
    }
}

/* Checks the connection table to find the key, if key is found, it grabs data,
   o/w it inserts the key to the connection table
   input: pointer to the key, pointer to the data, current time
   output: pointer to the data is set to proper place */

void insertKeyGrabDataFromConnTable(struct conn_key* ck, struct conn_data** cdata, double time)
{
  int location;
  int outcome;

  /* Find the index for the connection key in the table */ 
  location = find(ck,conntable);

  /* Check if connection key can be found and insert it to the table */
  if (location == NOT_FOUND)
    location = insert(ck, conntable, &outcome);
  
  /* Pointer to the entry in connection table is set here */
  *cdata = (struct conn_data*)conntable->entries[location].data;

  /* Check if the connection table is more than %90 percent full */
  if(conntable->FULL_ALERT)
    {
      emergencyCleanUpConnTable(location);
    }

  /* Check if connection is just inserted */
  if (outcome == NEW)
    {
      init_conn(*cdata);
    }
  (*cdata)->lasttimestamp = time;
}

/* Checks the destination table to find the key, if key is found, it graps the data,
   o/w it inserts the key to the destination table
   input: pointer to the key, pointer to the data, current time
   output: pointer to the data is set to the proper place */

void insertKeyGrabDataFromDestTable(struct dest_key* dk, struct dest_data** ddata, double time)
{
  int location;
  int outcome;

  /* Find the index for the destination key in the table */ 
  location = find(dk,desttable);

  /* Check if destination key can be found and insert it to the table */
  if (location == NOT_FOUND)
    location = insert(dk, desttable, &outcome);
  
  /* Pointer to the destination table entry is set here */
  *ddata = (struct dest_data*)desttable->entries[location].data;

  /* Check if the destination table is more than %90 percent full */
  if(desttable->FULL_ALERT)
    {
      emergencyCleanUpDestTable(location);
    }

  /* Check if destination is just inserted */
  if (outcome == NEW)
    init_dest(*ddata,time);

  (*ddata)->lasttimestamp = time;
}

/* Performs the mapping between the packet type and the index in the arrays
   input: packet type
   output: index for the packet type */

int gettype(enum ptype packettype)
{
  switch(packettype)
    {
    case TCP:
      return TCPI;
    case UDP:
      return UDPI;
    case ICMP:
      return ICMPI;
    default:
      return OTHERI;
    }
}

/* Adds the specified amount to the place that pointer points to in the array
   input: pointer to array element, the amount that will be added
   output: array is updated */

void insertToArray(int* array, int len)
{
  array[0] += len;
}

/* Determines which attacks are detected with more confidence than others 
   input: two indexes that will be compared
   output: non-zero value will be returned if the first index has more 
   confidence level than the second, o/w 0 will be returned */

int precedence(int i, int j)
{
  return (confidence[i] > confidence[j]);
}

/* Checks if the source and destination IP addresses are the same
   input: source IP address, destination IP addresses
   output: non-zero value will be returned if arguments are same
   o/w 0 will be returned */

int malformed(int source, int dest)
{
  return (source == dest);
}

/* Checks if the packet type is reserved packet type
   input: packet type
   output: non-zero value will be returned if packet type is reserved
   packet type, o/w 0 will be returned */

int non_existing(int packettype)
{
  return ((packettype >= LOW_PROTO)&&(packettype <= HIGH_PROTO));
}

/* Checks to see whether packet fits the attack detected, it is to eliminate
   the case where more than one attack type is set to TRUE, if the packet doesn't fit to
   attack characteristics that have highest precedence, it's not considered as valid.
   input: attack type, packet type, syn flag, ack flag, source and destination IP addresses
   fragmenation flag, no-flag-set flag
   output: FALSE returned if packet doesn't fit attack, non-zero value returned o/w */

int fittype(enum atype attacktype, enum ptype packettype, int syn, int ack,
	    unsigned int source, unsigned int dest, int frag, int noflag)
{
  switch(attacktype)
    {
    case TOO_MANY_MISMATCHES:
    case TCP_RTO_TOO_BIG:
    case NO_TCP_FLAG:
      return ((packettype == TCP) && noflag);
    case ICMP_FLOOD:
      return (packettype == ICMP);
    case UDP_FLOOD:
      return (packettype == UDP);
    case SYN_FLOOD:
      return (packettype == TCP && syn && !ack);
    case MALFORMED:
      return malformed(source,dest);
    case PRIVATE_IP:
      return (privateIP(source) || privateIP(dest));
    case TOO_MANY_FRAGS:
      return frag;
    case NONEXISTING_PROTO:
      return non_existing(packettype);
    default:
      return FALSE;
    }
}

/* Processes the packet and checks if it is malicous,updates our 
   connection and destination table
   input: time, packet type, source IP, destination IP
   source port number, destination port number, sequence number
   acknowledgement number, length of the packet, a variable that 
   shows if syn flag is set, a variable that shows if fin flag is set,
   a variable that shows if push flag is set, a variable that shows if 
   reset flag is set, a variable that shows if ack flag is set,a variable 
   that shows if packet is fragmented, packet header and the packet
   output: connection table is updated */

void process_packet(double time, int packettype, unsigned int source, 
		    unsigned int dest, unsigned short sport, 
		    unsigned short dport, unsigned int seqnum, 
		    unsigned int acknum, int len, 
		    int syn, int fin, int push, int reset, 
		    int ack, int fragmented, 
		    const struct pcap_pkthdr *hdr, const u_char* packet)
{
  /* Keys for the hash tables */
  struct dest_key dkey, skey;
  struct conn_key ckey, rckey;

  /* Pointers to entries in the hash tables */
  struct dest_data* ddata,* sdata;
  struct conn_data* cdata,* rcdata;

  enum atype attacktype = NO_ATTACK;
  enum atype victimtype[NO_OF_ATTACKS];

  int i, found = FALSE, j, noflag = FALSE;

  /* Set all attack flags to FALSE */
  for(i=0; i<NO_OF_ATTACKS; i++)
    victimtype[i] = FALSE;

  /* Compare prevous lines time to see whether we're in new second */
  if (time - resetTime >= 1 || first)
    {
      printf("New second %lf", time);

      resetCounters(desttable,conntable,time);
      resetTime = time;

      /* Set the first flag to FALSE */
      first = FALSE;
    }

  /* Set the keys accordingly */
  dkey.ip = dest;
  skey.ip = source;
  ckey.ip1 = source;
  ckey.ip2 = dest;
  ckey.port1 = sport;
  ckey.port2 = dport;
  rckey.ip1 = dest;
  rckey.ip2 = source;
  rckey.port1 = dport;
  rckey.port2 = sport;
  
  /* Insert records into destination table or find them in destination table */
  insertKeyGrabDataFromDestTable(&dkey, &ddata, time);
  insertKeyGrabDataFromDestTable(&skey, &sdata, time);
  
  /* Insert records into connection table or find them in connection table */
  insertKeyGrabDataFromConnTable(&ckey, &cdata, time);
  insertKeyGrabDataFromConnTable(&rckey, &rcdata, time);

  /* Update data statistics */
  ddata->bytesreceived[gettype(packettype)] += len;
  ddata->packetsreceived[gettype(packettype)] ++;
  sdata->bytessent[gettype(packettype)] += len;
  sdata->packetssent[gettype(packettype)] ++;
  insertToArray(ddata->bytesreceivedarray[gettype(packettype)], len);

  if (packettype == TCP)
    {

      /* Check for no flag case and detect this attack */
      if (!syn && !reset && !ack && !push && !fin)
	{
	  attacktype = NO_TCP_FLAG;
	  victimtype[NO_TCP_FLAG] = TRUE;
	  noflag = TRUE;
	  printf("No flag attack at time %lf ports %d %d\n", 
		 time, sport, dport);
	}

      /* Update related SYN packet ratios */
      if (syn)
	{
	  if (!ack)
	    {
	      sdata->synsent ++;
	      ddata->synreceived ++;
	    }
	  else
	    {	     
	      sdata->synacksent ++;
	      ddata->synackreceived ++;
	    }
	  ddata->synrto = ddata->synsentratio*ALPHA +
	    (double)ddata->synreceived
	    /(ddata->synacksent?ddata->synacksent:1)*(1-ALPHA);

	  /* Check for SYN flood */
	  if (ddata->synrto > MAX_SYN_RTO &&
	      ddata->synreceived - ddata->synacksent > SYN_SYNACK_DIFF)
	    {
	      attacktype = SYN_FLOOD;
	      victimtype[SYN_FLOOD] = TRUE;
	    }
	}

      /* Check here for TCP ratios and such */
      if (ddata->packetsreceived[TCPI] > MIN_PACKETS)
	{
	  ddata->tcprto = ddata->tcprto*ALPHA + (double)ddata->packetssent[TCPI] 
	    / (ddata->packetsreceived[TCPI]?ddata->packetsreceived[TCPI]:1) * (1-ALPHA);
	  
	  /* Check for TCP flood */
	  if (ddata->tcprto > TCP_MAX_RTO)
	    {
	      attacktype = TCP_RTO_TOO_BIG;
	      victimtype[TCP_RTO_TOO_BIG] = TRUE;
	    }
	}

      /* Check legitimate with sequence numbers */
      if (len)
	{
	  check_legitimate(cdata,seqnum,len);
	}
      
      /* Detect the illegitimate connection with mismatching sequence numbers */
      if (cdata->legitimate == FALSE)
	{
	  attacktype = TOO_MANY_MISMATCHES;
	  victimtype[TOO_MANY_MISMATCHES] = TRUE;
	  printf("Mismatches on connection port %d and %d "
		 "numnegatives %d seqnum %u\n",
		 sport, dport, cdata->oop, seqnum);
	} 
    }

  /* If packet is ICMP and enough traffic is seen */
  else if(packettype == ICMP && ddata->packetsreceived[ICMPI] > MIN_PACKETS)
    {
      ddata->icmprto = ddata->icmprto*ALPHA + (double)ddata->packetsreceived[ICMPI]
	/ ((ddata->packetssent[ICMPI] > 0)?ddata->packetssent[ICMPI]:1) * (1-ALPHA);
      
      /* Check for ICMP flood */
      if (ddata->icmprto > ICMP_MAX_RTO)
	{
	  struct in_addr in;
	  attacktype = ICMP_FLOOD;
	  victimtype[ICMP_FLOOD] = TRUE;
	  in.s_addr = dest;
	}
    }
  
  /* If packet is UDP and enough traffic is seen */
  else if (packettype == UDP && ddata->packetsreceived[UDPI] > MIN_PACKETS)
    {

      /* Calculate average rate of traffic in current and past window for TCP 
	 and UDP traffic */ 
      int curUDP = rateinwindow(ddata->bytesreceivedarray[gettype(packettype)], 0, WSIZE);
      int oldUDP = rateinwindow(ddata->bytesreceivedarray[gettype(packettype)], WSIZE, 2*WSIZE-1);
      int curTCP = rateinwindow(ddata->bytesreceivedarray[TCPI], 0, WSIZE-1);
      int oldTCP = rateinwindow(ddata->bytesreceivedarray[TCPI], WSIZE, 2*WSIZE -1);

      /* Calculate UDP received to sent ratio */
      double udprto = ddata->packetsreceived[UDPI]/(ddata->packetssent[UDPI]?ddata->packetssent[UDPI]:1);

      /* Check for UDP attacks. Raise alarm if UDP traffic doubled while TCP traffic
	 to the same destination subsided. Also check if there is a TCP connection between 
	 source and destination which indicates that the UDP stream is just media traffic. 
	 Finally, check if there are UDP replies from the victim.  */
      if (curUDP > 2*oldUDP && curTCP < 0.5*oldTCP && 
	  !findTCP(source,dest) && udprto > UDP_MAX_RTO)
	{
	  attacktype = UDP_FLOOD;
	  victimtype[UDP_FLOOD] = TRUE;

	  /* If this is the beginning of the attack, remember the attack rate
	     so we could later detect the end of the attack when the rate
	     goes below this value */
	  if (!ddata->victimtype[UDP_FLOOD])
	    ddata->stoprate = curUDP;
	} 

      /* Check if UDP FLOOD ended */
      if(curUDP < ddata->stoprate)
	{
	  ddata->victimtype[UDP_FLOOD] = FALSE;
	  victimtype[UDP_FLOOD] = FALSE;
	}
    }

  else
    {
      /* Check if packet has unknown protocol type*/
      if(non_existing(packettype))
	{
	  attacktype = NONEXISTING_PROTO;
	  victimtype[NONEXISTING_PROTO] = TRUE;
	}
    }
  
  /* Set the protocol type of the connection */
  cdata->protocol = packettype;
  rcdata->protocol = packettype;

  /* Check for reserved IP's */
  if (checkprivateIPflag && (privateIP(source) || privateIP(dest)))
    {
      attacktype = PRIVATE_IP;
      victimtype[PRIVATE_IP] = TRUE;
    }
  
  /* Update the number of packets sent */
  cdata->sent++;
  rcdata->received++;

  /* Check if packet is fragmented */
  if(fragmented)
    {
      double fragmentrate;

      /* Update number/rate of fragmented packets */
      cdata->nooffragmented++;
      fragmentrate = ((double)cdata->nooffragmented/cdata->sent);

      /* Detect a fragmentation attack */
      if((fragmentrate > MAX_FRAG_RATE)&&(cdata->sent > MIN_CONNECTION_LIMIT))
	{
	  attacktype = TOO_MANY_FRAGS;
	  victimtype[TOO_MANY_FRAGS] = TRUE;
	}
    }
 
  /* Detect the attack with highest precedence */
  for(i=1; i<NO_OF_ATTACKS; i++)
    {
      if (ddata->victimtype[i])
	{
	  int j;

	  found = TRUE;
	  for(j=1; j<NO_OF_ATTACKS; j++)
	    if (victimtype[j] && (j != i) && precedence(j, i))
	      {
		struct in_addr in;

		/* Set the alarm for attack with smaller precedence to FALSE */
		in.s_addr = htonl(dest);
		ddata->victimtype[i] = FALSE;
		fprintf(victimfile, "%lf %d %u %s\n", time, j, dest, inet_ntoa(in));
		printf("%lf Detected attack %d on %u noflag %d proto %d\n", time, j, dest, noflag, packettype);
		fflush(victimfile);
	      }
	}
    }

  if (!found && attacktype != NO_ATTACK)
    {   
      struct in_addr in;
      enum atype curtype = NO_ATTACK;

      in.s_addr = htonl(dest);

      for(j=1; j<NO_OF_ATTACKS; j++)
	if (victimtype[j] && precedence(j, curtype))
	  curtype = j;
      ddata->victimtype[curtype] = TRUE;
      fprintf(victimfile, "%lf %d %u %s\n", time, curtype, dest,
	      inet_ntoa(in));
      printf("%lf Just now detected attack %d on %u noflag %d proto %d\n", 
	     time, curtype, dest, noflag, packettype);
      fflush(victimfile);
    }

  found = FALSE;

  /* Check if the packet and the attack detected fit each other */
  for(i=1; i<NO_OF_ATTACKS; i++)
    {
      if (ddata->victimtype[i] && fittype(i, packettype, syn, ack, source, 
					  dest, fragmented, noflag))
	{ 
	  write_pcap(hdr, packet, 1);
	  found = TRUE;
	  break;
	}
    }
  if (!found)
    {
      write_pcap(hdr, packet, 0);    
    }
}

/* Explains the usage of the program
   input: name of the program
   output: a warning about using the program 
   and command line arguments are printed to the screen */

void usage(char* prg)
{
  printf("Usage %s tcpdump-input-file legitimate-output-file attack-output-file victim-file [check-reserved-IP-flag]\n", prg);
  exit(1);
}

int main(int argc, char** argv)
{
  /* Check if correct number of comand line arguments are given */
  if (argc < 5)
    usage(argv[0]);

  /* Check for the optional argument */
  if (argc > 5)
    checkprivateIPflag = atoi(argv[5]);

  /* Open file for attack alerts */
  victimfile = fopen(argv[4],"w");
  if (!victimfile)
    {
      printf("Could not open victim file %s\n", argv[4]);
      exit(1);
    }

  /* Hash tables are being created */
  create_hash(&desttable,DEST_TABLE_SIZE,DEST_TABLE_SIZE,DEST_TABLE_SIZE,1,0,
	      sizeof(struct dest_key),sizeof(struct dest_data),(int (*)(void *))dest_clean,(int (*)(void *,void *))dest_compare,
	      ( unsigned long(*)(void *, unsigned long ))dest_hash1,(unsigned long(*)(unsigned long, void*, unsigned long))dest_hash2);
  create_hash(&conntable,CONN_TABLE_SIZE,CONN_TABLE_SIZE,CONN_TABLE_SIZE,1,0,
	      sizeof(struct conn_key),sizeof(struct conn_data),(int (*)(void *))conn_clean,(int (*)(void *,void *))conn_compare,
	      ( unsigned long(*)(void  *,unsigned long ))conn_hash1,(unsigned long(*)(unsigned long, void*,unsigned long))conn_hash2);

  read_pcap(3, argv); 
  fflush(victimfile);
  fclose(victimfile);
  return 0;  
}











