/* This tool was developed at the University of Delaware and is publicly
   available for academic and non-commercial use. University of Delaware
   assumes no responsibility for consequences arising from the use of
   this tool. It is released under the GPL license. All questions and
   feedback should be directed to Jelena Mirkovic <sunshine@cis.udel.edu> */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#include "pcap_packet_read.h"

#include "customize.h"
#include "hash.h"
#include "constants.h"

#include "dest.h"
#include "conn.h"
#include "pair.h"

FILE * onewayfile;

struct hash_table* conntable;

int ports[NUM_PORTS];
int numports=0;

/* Explains the usage of the program
   input: name of the program
   output: a warning about using the program 
   and command line arguments are printed to the screen */

void usage(char* prg)
{
  printf("Usage %s tcpdump-input-file oneway-file forbidden-ports-file\n", prg);
  exit(1);
}

/* Reads in the list of forbidden ports. Traffic to/from IPs 
   that use these ports will be filtered out. 
   File must be sorted in ascending order.
   input: pointer to a file with port numbers one per line
   output: none
*/
void readports(FILE* portfile)
{
  while (TRUE)
    {
      fscanf(portfile, "%d", &ports[numports]);
      if (feof(portfile))
	break;
      numports++;
    }
  printf("Read %d forbidden ports\n", numports);
}

/* Finds a port in the list.
   input: port number
   output: found/not found
*/
int findport(int port)
{
  int i, high=numports, low=-1;
  do{
    i = (high+low)/2;
    if (ports[i] == port)
      return TRUE;
    if (ports[i] < port)
      low = i;
    else
      high = i;
  }while (high > low+1);
  return FALSE;
}

/* Closes the files that are previously opened.
   input: global file pointers
   output: files are closed */

void closefiles()
{
  fclose(onewayfile);
}


/* Initializes all the attributes in the connection table entry 
   input: pointer to connection table entry 
   output: initalized values of connection table entry */

void init_conn(struct conn_data * data)
{
  data->firstseqnum = INIT_SEQ_NUM;
  data->lastseqnum = INIT_SEQ_NUM;
  data->oop = 0;
  data->oneway = TRUE;
  data->legitimate = TRUE;
}



/* Checks the connection table to find the key, if key is found, it grabs data,
   o/w it inserts the key to the connection table
   input: pointer to the key, pointer to the data, current time
   output: pointer to the data points to right place */

void insertKeyGrabDataFromConnTable(struct conn_key* ck, struct conn_data** cdata, double time)
{
  int location;
  int outcome = EXISTING;

  /* Index for the connection in the table is going to be found here */ 
  location = find(ck,conntable);

  /* If we can't find the key in the connection table, than we're going 
     to insert the key for the connection to the connection table */
  if (location == NOT_FOUND)
    location = insert(ck, conntable, &outcome);
  
  /* Pointer to entry of the connection is set here */
  *cdata = (struct conn_data*)conntable->entries[location].data;

  if (outcome == NEW)
    {
      // printf("Initializing %d %d \n", ck->port1, ck->port2);
      init_conn(*cdata);
    }

  (*cdata)->lasttimestamp = time;
}

/* If the connection is oneway and legitimate TCP connection 
   with port numbers bigger than 0, connection is written to one-way file
   and deleted from the connection table
   input: pointer to the key, pointer to the entry in the connection table 
   output: connection is written to the oneway file and is deleted */

void closeConnection(struct conn_key* ckey, struct conn_data* cdata)
{
  if (cdata->oneway && cdata->legitimate && cdata->sent > MIN_PACKETS 
      && cdata->protocol == TCP && ckey->port1 > 0 && ckey->port2 > 0)
    fprintf(onewayfile, "%u %d %u %d\n", ckey->ip1, ckey->port1, ckey->ip2, ckey->port2);

  delete(ckey, conntable);
}

/* Legitimacy of connection is checked, sequence numbers, suspicious points, 
   number of retransmissions for the connection are updated.
   input: pointer to the entry in the connection table, sequence number of
   the packet, length of the packet
   output: sequence numbers,suspicious points, number of retransmissions
   for the connection are updated, legitimacy of the connection is checked */


void check_legitimate(struct conn_data* cdata, unsigned int seqnum, int len)
{
  if (cdata->legitimate == FALSE)
    return;

  /* Update seqnum's for the connection */
  if(seqnum == cdata->lastseqnum || cdata->firstseqnum == INIT_SEQ_NUM)
    {
      /* Update seqnum's for the connection */
      cdata->firstseqnum = seqnum;
      cdata->lastseqnum = seqnum + len;
    }

  /* Check for same packets */
  else if (seqnum == cdata->firstseqnum && seqnum+len == cdata->lastseqnum)
      {
	cdata->oop -= SEQ_DECREMENT;
	cdata->retransmissions++;
      }

  /* Not same but out of order */
    else
      {
	cdata->oop += (seqnum - cdata->lastseqnum);
	if (seqnum > cdata->lastseqnum)
	  {
	    cdata->firstseqnum = seqnum;
	    cdata->lastseqnum = seqnum+len;
	  }
      }

  /* If sum of suspicous points are bigger than the threshold, connection is marked as non-legitimate  */
  if(abs(cdata->oop) > MAX_SEQ_DIFF)
    {
      cdata->legitimate = FALSE;
    }

  /* If number of retransmissions exceeds the threshold, connection is marked as non-legitimate */
  if (cdata->retransmissions > MAX_RETRANSMISSIONS)
    {
      cdata->legitimate = FALSE;
    }
}

/* Processes the packet and updates our connection table
   input: time, packet type, source IP, destination IP
   source port number, destination port number, sequence number
   acknowledgement number, length of the packet, a variable that 
   shows if syn flag is set, a variable that shows if fin flag is set,
   a variable that shows if push flag is set, a variable that shows if 
   reset flag is set, a variable that shows if ack flag is set,a variable 
   that shows if packet is fragmented, packet header and the packet
   output: connection table is updated */

void process_packet(double time, enum ptype packettype, unsigned int source, 
		    unsigned int dest, unsigned short sport, 
		    unsigned short dport, unsigned int seqnum, 
		    unsigned int acknum, int len, int syn, int fin, 
		    int push, int reset, int ack, int fragmented,  
		    const struct pcap_pkthdr *hdr, const u_char* packet)
{
  int outcome;
  
  /* Connection table keys */
  struct conn_key ckey, rckey;
  struct conn_data* cdata,* rcdata;
  struct pair_key pkey,rpkey;

  /* If this data is to/from a port on a list of forbidden ports just
     remember source/dest pairs as oneway */
  if (findport(sport) || findport(dport))
    {
      printf("Found port %d or %d\n", sport, dport);
      fprintf(onewayfile, "%u %d %u %d\n", source, sport, dest, dport);
      fflush(onewayfile);
      return;
    }
  /* Initialize keys */
  ckey.ip1 = source;
  ckey.ip2 = dest;
  ckey.port1 = sport;
  ckey.port2 = dport;
  rckey.ip1 = dest;
  rckey.ip2 = source;
  rckey.port1 = dport;
  rckey.port2 = sport;
  pkey.ip1 = source;
  pkey.ip2 = dest;
  rpkey.ip1 = dest;
  rpkey.ip2 = source;
 
  /* If data exists in the connection table we find it, if not we insert our keys */
  insertKeyGrabDataFromConnTable(&ckey, &cdata, time);
  insertKeyGrabDataFromConnTable(&rckey, &rcdata, time);
  
  cdata->protocol = packettype;
  rcdata->protocol = packettype;

  /* Remove the one way flag if we saw some traffic in the opposite direction */
  if(rcdata->oneway == TRUE)
    {
      rcdata->oneway = FALSE;
    }

  cdata->sent++;
  rcdata->received++;
  
  if (packettype != TCP)
    return;
  
  /* For fin and reset flags, we close the connection */
  if (fin || reset)
    closeConnection(&ckey, cdata);

  if (!push)
    return;

  /* If non of the flags are set, it's non-legitimate traffic */
  if (!syn && !reset && !ack && !push && !fin)
    cdata->legitimate = FALSE;
  
  /*  if(!syn && ack)
      cdata->legitimate =TRUE; */
  
  /* If connection seems to be oneway, legitimacy of connection is checked */
  if(cdata->oneway)
    check_legitimate(cdata, seqnum, len);
}


int main(int argc, char** argv)
{
  unsigned int i;
  FILE* portsfile;

  if (argc != 4)
    usage(argv[0]);

  /* Oneway file is opened properly */
  onewayfile = fopen(argv[2],"w");
  if(onewayfile == NULL)
    {
      printf("One-way file %s can not be opened!\n", argv[2]);
      exit(1);
    }

  /* Open the forbidden ports file and read in data */
  portsfile = fopen(argv[3],"r");
  if(portsfile == NULL)
    {
      printf("Forbidden ports file %s can not be opened!\n", argv[3]);
      exit(1);
    }

  readports(portsfile);

  /* Hash tables are being created */
  create_hash(&conntable,CONN_TABLE_SIZE,CONN_TABLE_SIZE,CONN_TABLE_SIZE,1,0,
	      sizeof(struct conn_key),sizeof(struct conn_data),(int (*)(void *))conn_clean,(int (*)(void *,void *))conn_compare,
	      ( unsigned long(*)(void  *,unsigned long ))conn_hash1,(unsigned long(*)(unsigned long, void*,unsigned long))conn_hash2);


  /* Packets are being read */
  read_pcap(1, argv); 

  /* Clean up connection table */  
  for(i = 0;i < conntable->size;i++)
    {
      if (conntable->entries[i].state == VALID)
	{
	  struct conn_key* ckey = (struct conn_key *)conntable->entries[i].key;
	  struct conn_data* cdata = (struct conn_data *)conntable->entries[i].data;
	  closeConnection(ckey, cdata);
	}
    }
}











