/* This tool was developed at the University of Delaware and is publicly
   available for academic and non-commercial use. University of Delaware
   assumes no responsibility for consequences arising from the use of
   this tool. It is released under the GPL license. All questions and
   feedback should be directed to Jelena Mirkovic <sunshine@cis.udel.edu> */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h> 
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>

#include "customize.h"
#include "hash.h"
#include "constants.h"

#include "sdest.h"
#include "conn.h"

#include "trace.h"
#include "pcap_packet_read.h"

FILE* victimfile_in, *human, *machine;
double lasttime;

struct hash_table *desttable;

struct trace* attack_trace[MAX_VICTIMS];

char attackname[NO_OF_ATTACKS][MAX_NAME]=
  {"No Attack", "Seqnum mismatch", 
   "TCP-ratio large", "SYN flood",
   "Malformed packet", "Non-exisiting protocol",
   "Private IP", "UDP flood", "ICMP flood", 
   "Fragmented traffic", "TCP no-flag"};

char protocolname[NUM_PROTOCOLS][MAX_NAME]=
  {"TCP","UDP", "ICMP","OTHER"};

char spoofingname[NUM_SPOOFING][MAX_NAME]=
  {"NO_SPOOFING", "RANDOM_SPOOFING", "SUBNET_SPOOFING", "PRIVATE_SPOOFING"};

struct alert alerts[MAX_ALERTS];
int first_alert=0;
int last_alert=-1;
int num_alerts=0;

/* Checks if the source and destination IP addresses are the same
   input: source IP address, destination IP addresses
   output: non-zero value will be returned if arguments are same
   o/w 0 will be returned */

int malformed(int source, int dest)
{
  return (source == dest);
}

/* Checks if the packet type is reserved packet type
   input: packet type
   output: non-zero value will be returned if packet type is reserved
   packet type, o/w 0 will be returned */

int non_existing(int packettype)
{
  return ((packettype >= LOW_PROTO)&&(packettype <= HIGH_PROTO));
}

/* Translates the protocol number to protocol name
   input: protocol number
   output: name of the protocol */

char* getProtocol(int protocol)
{
  switch(protocol)
    {
    case TCP:
      return protocolname[TCPI];
    case UDP:
      return protocolname[UDPI];
    case ICMP:
      return protocolname[ICMPI];
    default:
      return protocolname[OTHERI];
    }
}

/* Tests if IP address is private address
   input: IP address in decimal form
   output: returns TRUE if IP address is reserved, o/w returns FALSE */

int privateIP(unsigned int ip)
{
  if ((ip >> 24) == 10)
    return TRUE;
  else if((ip >> 24) == 172)
    {
      int second = ((ip << 8) >> 24);
      if((second >= 16) && (second <= 31))
        return TRUE;
      else
        return FALSE;
    }
  else if((ip >> 24) == 192)
    {
      int second = ((ip << 8) >> 24);
      if(second == 168)
        return TRUE;
      else
        return FALSE;
    }
  else
    return FALSE;
}

/* Creates a new trace to the attack trace list and initializes it
   input: index for the new trace in the list, IP address for the 
   victim, a parameter that shows if packets in trace has no flag set ,
   packet type,a parameter that shows if the packets in the trace has syn flag on,
   a parameter that shows if they have same source and destination IP and a parameter 
   that shows if it is fragmented
   output: trace is created and initialized */

void make_new_trace(int free, unsigned int dest,  int noflag, 
		    int packettype, int syn, int isMalformed, int frag)
{
  int i;

  attack_trace[free] = (struct trace*)malloc(sizeof(struct trace));
  attack_trace[free]->victim = dest;
  attack_trace[free]->packets = (struct packet*)
    malloc(INIT_NP*sizeof(struct packet));
  attack_trace[free]->limit = INIT_NP;
  attack_trace[free]->np = 0;
  attack_trace[free]->noflag = noflag;
  attack_trace[free]->proto = packettype;
  attack_trace[free]->syn = syn;
  attack_trace[free]->malformed = isMalformed;
  attack_trace[free]->frag = frag;
  printf("New victim %u\n",dest);
}

/* Inserts the new packet into the packet list of the trace 
   input: pointer to the trace, packet header, packet, time, 
   protocol number
   output: packet is inserted */

void insert_packet(struct trace* t, const struct pcap_pkthdr* hdr, 
		   const u_char* packet, double time, int proto)
{
  if (t->np == 0)
    t->start_time = time;
  t->packets[t->np].hdr = (struct pcap_pkthdr*)
    malloc(sizeof(struct pcap_pkthdr));
  memcpy(t->packets[t->np].hdr, hdr, sizeof(struct pcap_pkthdr));
  t->packets[t->np].packet = (unsigned char*)
    malloc(hdr->caplen);
  memcpy(t->packets[t->np].packet, packet, hdr->caplen);
  t->last_time = time;
  t->np++;
  /* Here check how to enlarge arrays */
  if (t->np == t->limit)
    {
      struct packet* temp =  (struct packet*)
	malloc(2*t->limit*sizeof(struct packet));
      if (!temp)
	exit(1);
      memcpy(temp, t->packets, t->limit*sizeof(struct packet));
      t->limit = 2*t->limit;
      free(t->packets);
      t->packets = temp;
      printf("enlarged array for %x to %d\n", t->victim, t->limit);
    }
}

/* Counts the number of ones in the binary form of the parameter 
   input: number whose binary form will be examined
   output: number of ones in the binary form of the parameter 
   will be returned */

int countOnes(unsigned char num)
{
  int count = 0;
  unsigned char num1;
  do
    {
      num1 = num & (num-1);
      if (num != num1)
	count++;
      num = num1;
    } while(num);
  return count;
}

/* Insert a destination record into a sorted array by address */
int insertSort(struct sdest_data** array, struct sdest_data* elem, int size)
{
  int low = -1, high = size, mid, j;
  if (size == 0)
    {
      array[0] = elem;
      return TRUE;
    }
  do
    {
      mid = (high+low)/2;
      if (!array[mid])
	break;
      if (array[mid]->IP == elem->IP)
	return FALSE;
      if (array[mid]->IP < elem->IP)
	low = mid;
      else
	high = mid;
    }
  while (high > low + 1);
  mid = (high+low)/2;
  if (!array[mid])
    {
      array[mid] = elem;
      return TRUE;
    }
  else 
    {
      for (j=size; j>mid; j--)
	{
	  array[j] = array[j-1];
	}
      array[mid] = elem;
      return TRUE;
    }
}

/* Find a destination record in a sorted array by address */
int findSort(struct sdest_data** array, unsigned int IP, int size)
{
  int low = -1, high = size, mid, j;
  if (size == 0)
    {
      return NOT_FOUND;
    }
  do
    {
      mid = (high+low)/2;
      if (array[mid]->IP == IP)
	return mid;
      if (array[mid]->IP < IP)
	low = mid;
      else
	high = mid;
    }
  while (high > low + 1);
  mid = (high+low)/2;
  if (array[mid]->IP == IP)
    return mid;
  return NOT_FOUND;
}



/* Finds the alert for the trace, gets attack attributes,  
   detects spoofing type by processing the trace , prints traces and extracted
   information to files
   input: pointer to the trace
   output: human and machine files are written */

void printTrace(struct trace* t)
{
  double time = t->start_time;
  double min = 10;
  int imin = NOT_FOUND;
  struct in_addr in;
  int i, j;
  int sum = 0;
  double duration = t->last_time - t->start_time;
  struct alert temp[MAX_ALERTS];
  int temp_num_alerts = 0;
  int num_reserved = 0;
  int k = 0, n;
  enum spooftype spoofing;
  int num_sources = 0, real_sources = 0, real_packets = 0;
  double Ehit = 0, phit = 1-pow(1-1.0/256,t->np);

  unsigned char sourcearray[NUM_OCTETS][NUM_ADDRESSES];
  int octet[NUM_OCTETS];
  unsigned char sport_array[NUM_PORTS];
  unsigned char dport_array[NUM_PORTS];
  int packets[NUM_PORTS];
  int sports=0, dports=0;
  struct sdest_data* print_sources[MAX_SOURCES];

  in.s_addr = htonl(t->victim);

  /* Initialize arrays */
  for(i = 0; i < NUM_OCTETS; i++)
    for(j=0; j < NUM_ADDRESSES; j++)
      sourcearray[i][j] = 0;
  for(i = 0; i < NUM_PORTS; i++)
    {
      sport_array[i] = 0;
      dport_array[i] = 0;
      packets[i] = 0;
    }
  for(i = 0; i < MAX_SOURCES; i++)
    print_sources[i] = 0;

  /* Clean the hash table */
  delete_all(desttable);
  
  /* Find the alert for this trace */
  for(i=first_alert;;)
    {
      if(fabs(alerts[i].time - time) < min && 
	 fitAlert(t, alerts[i].attacktype, alerts[i].victim))
	{
	  min = fabs(alerts[i].time - time);
	  imin = i;
	}
      i = (i+1)%MAX_ALERTS;
      if (i == (last_alert+1)%MAX_ALERTS)
	break;
    }
  if (imin != NOT_FOUND)
    {
      printf("Found alert for trace proto %d flag %d syn %d frag %d victim %u time %lf packets %d\n",
	     t->proto, t->noflag, t->syn, t->frag, t->victim, time, t->np);
    }
  
  /* Only print alerts for which we have enough packets 
     to make a decision */
  if (t->np > MIN_TRACE_PACKETS && imin > NOT_FOUND)
    {
      int rl = FALSE;
      double aggressive = 0;
      /* Print packets */
      for(i=0;i<t->np;i++)
	{
	  /* Get packet informaton */
	  unsigned char* packet = t->packets[i].packet;
	  struct pcap_pkthdr* hdr = t->packets[i].hdr;
	  double dtime = hdr->ts.tv_sec + (double)hdr->ts.tv_usec/1000000; 
	  struct ip* ip = (struct ip*)(packet + sizeof(struct ether_header));
	  int len = ntohs(ip->ip_len);
	  unsigned int srcIP = ntohl(ip->ip_src.s_addr);
	  unsigned int dstIP = ntohl(ip->ip_dst.s_addr);
	  char source[MAX_ADDRESS], dest[MAX_ADDRESS];
	  int proto = ip->ip_p;
	  struct in_addr in;
	  int frag = (ntohs(ip->ip_off) & 8192) || ((ntohs(ip->ip_off) & 8191) > 0);
	  int isfragment = ((ntohs(ip->ip_off) & 8191) > 0);
	  int sport = 0, dport = 0, outcome, index;
	  struct sdest_key dk;
	  struct sdest_data* ddata;


	  /* Avoid printing too short packets that are probably some weird
	     ethernet packets and not IP */
	  if (hdr->caplen < sizeof(struct ip) + sizeof(struct ether_header))
	    continue;
	  sum += len;
	  in.s_addr = htonl(srcIP);

	  strcpy(source, inet_ntoa(in));
	  
	  in.s_addr = htonl(dstIP);
	  strcpy(dest, inet_ntoa(in));
	  
	  /* Insert the source subnet into hash table */
	  dk.ip = srcIP;
	  index = find(&dk, desttable);
	  if (index == NOT_FOUND)
	    {
	      int pos;

	      index = insert(&dk, desttable, &outcome);
	      if (index == NOT_FOUND)
		{
		  printf("Alert! Some sources will not be counted because we ran out of memory!\n");
		}
	      else
		{
		  /* Initialize source count */
		  ddata = (struct sdest_data*)desttable->entries[index].data;
		  ddata->IP = srcIP;
		  /* Initialize ttl and drift */
		  ddata->ttl = ip->ip_ttl;
		  ddata->drift = 0;
		  ddata->packets = 0;
		  ddata->start = dtime;
		  ddata->isAg = FALSE;
		  ddata->similarity = 0;
		}
	    }
	  
	  /* If it is found in the table */
	  if (index != NOT_FOUND)
	    {
	      /* Find the position in the subnet */
	      ddata = (struct sdest_data*)desttable->entries[index].data;
	      
	      /* Calculate drift */
	      if (ip->ip_ttl == ddata->ttl - 1)
		{
		  ddata->drift ++;
		  ddata->ttl--;
		}
	      else if (ddata->ttl == 1)
		ddata->ttl = ip->ip_ttl;
	      /* Increase packet count */
	      ddata->packets++;
	      ddata->stop = dtime;
	      if (ddata->packets > MIN_PPS)
		{
		  struct in_addr sin;
		  double pps = ddata->packets/(ddata->stop - ddata->start);
		  sin.s_addr = srcIP;
		  if (proto == IPPROTO_TCP && pps > MAX_TCP_PPS)
		    {
		      if (!ddata->isAg)
			{
			  printf("Source %s aggressive, TCP pps is %lf\n", inet_ntoa(sin), pps);
			  ddata->isAg = TRUE;
			  aggressive += pps;
			}
		    }
		  else if (proto == IPPROTO_ICMP && pps > MAX_ICMP_PPS)
		    if (!ddata->isAg)
		      {
			printf("Source %s aggressive, ICMP pps is %lf\n", inet_ntoa(sin), pps);
			ddata->isAg = TRUE;
			aggressive += pps;
		      }
		}
	      if (ddata->drift > MAX_DRIFT)
		{
		  struct in_addr vin;
		  vin.s_addr = ntohl(dstIP);
		  rl = TRUE;
		  printf("RL is true for victim %s drift %d\n", inet_ntoa(vin), ddata->drift);
		  //break;
		}
	    }
	  
	  /* Check for reserved source IP */
	  if(privateIP(srcIP) == TRUE)
	    {
	      num_reserved++;
	    }

	  /* Get octets from source IP */
	  for(k=24, j=0; k >= 0; k -= 8, j++)
	    {
	      octet[j] = ((srcIP >> k) & 0xff);
	    }

	  /* Saturating counter */
	  for(k=0; k < NUM_OCTETS; k++)
	    if (sourcearray[k][octet[k]] < 255)
	      sourcearray[k][octet[k]] = sourcearray[k][octet[k]] ++;
	  
	  /* For TCP print following information to human file*/
	  if (proto == IPPROTO_TCP)
	    {
	      struct tcphdr* tcp;
	      unsigned int seq, ack;

	      tcp = (struct tcphdr*)((unsigned int)ip+4*ip->ip_hl);
	      sport = ntohs(tcp->source);
	      dport = ntohs(tcp->dest);
	      seq = ntohl(tcp->seq);
	      ack = ntohl(tcp->ack_seq);
	    }
	  /* For UDP print following information to human file */
	  else if (proto == IPPROTO_UDP)
	    {
	      struct udphdr* udp;

	      udp = (struct udphdr*)((unsigned int)ip+4*ip->ip_hl);
	      sport = ntohs(udp->source);
	      dport = ntohs(udp->dest);
	    }
	  
	  /* Update port statistics */
	  if (sport)
	    if (sport_array[sport] < 255)
	      sport_array[sport] = (sport_array[sport] + 1 ) % 256;
	  if (dport)
	    {
	      if (dport_array[dport] < 255)
		dport_array[dport] = (dport_array[dport] + 1 ) % 256;
	      packets[dport]++;
	    }
	}
      if (!rl) 
	{
	  /* Now decide if we have some type of spoofing */
	  for(i=0;i<NUM_OCTETS;i++)
	    {
	      octet[i] = 0;
	      for(j=0;j<NUM_ADDRESSES;j++)
		{
		  if (sourcearray[i][j] > 0)
		    octet[i]++;
		}
	    }
	  
	  /* Find the expected number of slots hit by randomly
	     spoofed packets given their number */
	  Ehit = NUM_ADDRESSES*(1-pow(1-phit,NUM_ADDRESSES))-1;
	  /* Detect subnet spoofing */
	  if(Ehit && octet[0] <= Ehit/10 && 
	     octet[1] <= Ehit/10 && 
	     octet[2] <= Ehit/10 && 
	     octet[3] > Ehit)
	    {
	      printf("Detected subnet spoofing hits %d %d %d %d ehit %lf\n", 
		     octet[0], octet[1], octet[2], octet[3], Ehit);
	      spoofing = SUBNET_SPOOFING;
	    }
	  else
	    /* Detect random spoofing */
	    if(Ehit && (octet[0] > Ehit || 
			octet[2] > Ehit || 
			octet[2] > Ehit) 
	       && octet[3] > Ehit)
	      spoofing = RANDOM_SPOOFING;
	    else 
	      /* Detect reserved spoofing */
	      if (num_reserved)
		spoofing = PRIVATE_SPOOFING;
	      else
		spoofing = NO_SPOOFING;
	  
	  /* Count the number of sources */
	  num_sources = 0;
	  
	  /* For all the subnets in the table */
	  for(k = 0; k < desttable->size; k++)
	    {
	      if (desttable->entries[k].state == VALID)
		{
		  struct sdest_data* ddata = (struct sdest_data*)
		    desttable->entries[k].data;		 
		  if (num_sources < MAX_SOURCES) 
		    insertSort(print_sources, ddata, num_sources);
		  num_sources ++;
		}
	    }
	  /* Calculate similarity first */
	  if (spoofing != RANDOM_SPOOFING)
	    {
	      double similarity = 0;
	      for(i=0;i<num_sources;i++)
		{
		  struct sdest_data* ddata = print_sources[i];
		  if (ddata->stop-ddata->start < MIN_DURATION ||
		      ddata->packets < MIN_PPH)
		    continue;
		  for(j=0;j<num_sources;j++)
		    {
		      double s, start;
		      if (i==j)
			continue;
		      struct sdest_data* odata = print_sources[j];
		      if (odata->stop-odata->start < MIN_DURATION || 
			  odata->packets < MIN_PPH)
			continue;
		      start = 1/fabs(ddata->start-odata->start);
		      if (start > 1)
			start = 1;
		      s = (ddata->stop-ddata->start)/
			(odata->stop-odata->start)*start;
		      if (s>1)
			s = 1/s;
		      if (s > similarity)
			similarity = s;
		    }
		  ddata->similarity = similarity;
		}
	    }
	  /* Now go through the packets and recalculate statistics for 
	     sources that matter */
	  if (spoofing != RANDOM_SPOOFING)
	    {
	      /* Clear port statistics */
	      for(i = 0; i < NUM_PORTS; i++)
		{
		  sport_array[i] = 0;
		  dport_array[i] = 0;
		  packets[i] = 0;
		}
	      /* Clear byte statistics */
	      sum = 0;
	      /* Go through packets */
	      for(i=0;i<t->np;i++)
		{
		  /* Get packet informaton */
		  unsigned char* packet = t->packets[i].packet;
		  struct pcap_pkthdr* hdr = t->packets[i].hdr;
		  double dtime = hdr->ts.tv_sec + 
		    (double)hdr->ts.tv_usec/1000000; 
		  struct ip* ip = (struct ip*)(packet + 
					       sizeof(struct ether_header));
		  int len = ntohs(ip->ip_len);
		  unsigned int srcIP = ntohl(ip->ip_src.s_addr);
		  unsigned int dstIP = ntohl(ip->ip_dst.s_addr);
		  int proto = ip->ip_p;
		  struct in_addr in;
		  int frag = (ntohs(ip->ip_off) & 8192) || 
		    ((ntohs(ip->ip_off) & 8191) > 0);
		  int isfragment = ((ntohs(ip->ip_off) & 8191) > 0);
		  int sport = 0, dport = 0, outcome, index;
		  struct sdest_key dk;
		  struct sdest_data* ddata;

		  /* Avoid printing too short packets that are
		     probably some weird
		     ethernet packets and not IP */
		  if (hdr->caplen < sizeof(struct ip) + 
		      sizeof(struct ether_header))
		    continue;
	
		  index = findSort(print_sources, srcIP,  num_sources);
		  /* Find the source in the array */
		  if (index == NOT_FOUND)
		    continue;
		  ddata = (struct sdest_data*)print_sources[index];
		  if ((ddata->similarity < SIM_THRESH || 
		      ddata->packets < MIN_PPH) && num_sources > 1)
		    continue;
		  sum += len;
		  real_packets ++;		 

		  if (proto == IPPROTO_TCP)
		    {
		      struct tcphdr* tcp;
		      unsigned int seq, ack;
		      
		      tcp = (struct tcphdr*)((unsigned int)ip+4*ip->ip_hl);
		      sport = ntohs(tcp->source);
		      dport = ntohs(tcp->dest);
		      seq = ntohl(tcp->seq);
		      ack = ntohl(tcp->ack_seq);
		    }
		  /* For UDP print following information to human file */
		  else if (proto == IPPROTO_UDP)
		    {
		      struct udphdr* udp;
		      
		      udp = (struct udphdr*)((unsigned int)ip+4*ip->ip_hl);
		      sport = ntohs(udp->source);
		      dport = ntohs(udp->dest);
		    }
		  
		  /* Update port statistics */
		  if (sport)
		    if (sport_array[sport] < 255)
		      sport_array[sport] = (sport_array[sport] + 1 ) % 256;
		  if (dport)
		    if (dport_array[dport] < 255)
		      dport_array[dport] = (dport_array[dport] + 1 ) % 256;
		  packets[dport]++;
		}
	      if (num_sources > 1)
		for(i=0;i<num_sources;i++)
		  {
		    struct sdest_data* ddata = print_sources[i];
		    if (ddata->similarity > SIM_THRESH && 
		      ddata->packets > MIN_PPH)
		      real_sources ++;	  
		  }
	      else
		real_sources = 1;
	    }
	  else
	    {
	      real_sources = num_sources;
	      real_packets = t->np;
	    }
	  /* Count ports */
	  for(i=0; i<NUM_PORTS; i++)
	    {
	      if (sport_array[i])
		{
		  sports++;
		}
	      if (dport_array[i])
		{
		  dports++;
		}
	    }
	  /* Now if we have no sources that count as attack give up */

	  if (real_sources > 0 && real_packets > MIN_TRACE_PACKETS)
	    {
	      if ((alerts[imin].attacktype == TOO_MANY_MISMATCHES ||
		   alerts[imin].attacktype == TCP_RTO_TOO_BIG ||
		   alerts[imin].attacktype == SYN_FLOOD ) && 
		  aggressive/real_sources < MAX_TCP_PPS 
		  && spoofing != RANDOM_SPOOFING)
		{
		  printf("Not aggressive %lf sources %d for victim %s\n", aggressive, num_sources, inet_ntoa(in));
		}
	      else if (alerts[imin].attacktype == ICMP_FLOOD && 
		       aggressive/real_sources < MAX_ICMP_PPS
		       && spoofing != RANDOM_SPOOFING)
		{
		  printf("Not aggressive %lf sources %d for victim %s\n", aggressive, num_sources, inet_ntoa(in));
		}
	      else
		{
		  /* Print to files */
		  fprintf(machine, "start = %lf victim = %s attack = %s duration = %lf pps = %lf Bps = %lf total packets = %d spoofing = %s sources = %d "
			  "source ports = %d dest ports = %d\n",
			  alerts[imin].time, inet_ntoa(in), 
			  attackname[alerts[imin].attacktype], 
			  duration, real_packets/duration, 
			  sum/duration, real_packets,
			  spoofingname[spoofing], real_sources, sports, dports);
		  fprintf(human, "start = %lf victim = %s attack = %s duration = %lf pps = %lf Bps = %lf total packets = %d spoofing = %s sources = %d "
			  "source ports = %d dest ports = %d\n",
			  alerts[imin].time, inet_ntoa(in), 
			  attackname[alerts[imin].attacktype], 
			  duration, real_packets/duration, 
			  sum/duration, real_packets,
			  spoofingname[spoofing], real_sources, sports, dports);
		  fprintf(human, "Sources: \n");
		  for(i=0;i<num_sources;i++)
		    {
		      struct sdest_data* ddata = print_sources[i];
		      in.s_addr = htonl(ddata->IP);
		      if ((spoofing == RANDOM_SPOOFING) || 
			  (num_sources == 1) ||
			  (ddata->packets > MIN_PPH && 
			   ddata->similarity > SIM_THRESH))
			{
			  fprintf(human, "IP = %s start = %lf stop = %lf active time = %lf packets sent = %d similarity = %lf\n", 
				  inet_ntoa(in), ddata->start,
				  ddata->stop, ddata->stop-ddata->start, 
				  ddata->packets, 
				  ddata->similarity);
			}
		    }
		  fprintf(human, "Source ports: \n");
		  for(i=0; i<NUM_PORTS; i++)
		    {
		      if (sport_array[i])
			{
			  fprintf(human, "%d\n", i);
			}
		    }
		  fprintf(human, "Dest ports: \n");
		  for(i=0; i<NUM_PORTS; i++)
		    {
		      if (dport_array[i])
			{
			  fprintf(human, "%d %d\n", i, packets[i]);
			}
		    }
		  fprintf(human,"-------------------------------------------------------\n");
		  fflush(human);
		}
	    }
	}
    }
  /* Now remove alerts we used */
  for(i=first_alert;;)
    {
      if (!(alerts[i].time >= t->start_time &&
	  alerts[i].time <= t->last_time &&
	    fitAlert(t, alerts[i].attacktype, alerts[i].victim)))
	{
	  temp[temp_num_alerts++] = alerts[i];
	}
      i = (i+1)%MAX_ALERTS;
      if (i == (last_alert+1)%MAX_ALERTS)
	break;
    }
  memcpy(alerts, temp, temp_num_alerts*sizeof(struct alert));
  first_alert = 0;
  last_alert = temp_num_alerts-1;
  num_alerts = temp_num_alerts;
}

/* Prints the trace to files and frees allocated memory
   input: pointer to the trace
   output: files are updated and memory is deallocated */

void closeTrace(struct trace* t)
{
  int i;

  printf("DONE\n");
  fflush(stdout);
  printTrace(t);
  for(i=0;i<t->np;i++)
    {
      free(t->packets[i].hdr);
      free(t->packets[i].packet);
    }
  free(t->packets);
}

/* Checks if the trace and attributes match
   input: pointer to the trace,IP address for the victim a parameter that 
   shows if packets in trace have no flag set , packet type,a parameter that 
   shows if the packets in the trace has syn flag on,a parameter that shows if
   they have same source and destination IP and a parameter that shows if it is 
   fragmented
   output: Non zero if they match, o/w 0 */

int isMatch(struct trace* t,  unsigned int  dest, int noflag, int packettype, 
	    int syn, int isMalformed, int frag)
{
  return ((t->victim == dest) && (t->noflag == noflag) && (t->proto == packettype) &&
	  (t->syn == syn) && (t->malformed == isMalformed) && (t->frag == frag));
}

/* Checks the correspondance between attacktype and victim with the trace information
   input: pointer to the trace, attack type, victim IP
   output: TRUE if it fits, o/w returns FALSE */

int fitAlert(struct trace* t, enum atype attacktype, unsigned int victim)
{
  if (t->victim != victim)
    {
      return FALSE;
    }
  switch(attacktype)
    {
    case TOO_MANY_MISMATCHES:
    case TCP_RTO_TOO_BIG:
      return (t->proto == TCP);
    case NO_TCP_FLAG:
      return ((t->proto == TCP) && t->noflag);
    case ICMP_FLOOD:
      return (t->proto == ICMP);
    case UDP_FLOOD:
      return (t->proto == UDP);
    case SYN_FLOOD:
      return (t->proto == TCP && t->syn);
    case MALFORMED:
      return (t->malformed);
    case PRIVATE_IP:
      return FALSE;
    case TOO_MANY_FRAGS:
      return t->frag;
    case NONEXISTING_PROTO:
      return non_existing(t->proto);
    default:
      return FALSE;
    }

}

/* Processes the packet, if alert list is out of date
   reads some more from victim file to alert list besides that
   searches for the trace of the packet in the trace list, if it
   can't find the trace, it creates a new trace
   input: time, packet type, source IP, destination IP
   source port number, destination port number, sequence number
   acknowledgement number, length of the packet, a variable that 
   shows if syn flag is set, a variable that shows if fin flag is set,
   a variable that shows if push flag is set, a variable that shows if 
   reset flag is set, a variable that shows if ack flag is set,a variable 
   that shows if packet is fragmented, packet header and the packet
   output: alert list and victim list are updated */

void process_packet(double time, int packettype, unsigned int source, 
		    unsigned int dest, unsigned short sport, 
		    unsigned short dport, unsigned int seqnum, 
		    unsigned int acknum, int len, 
		    int syn, int fin, int push, int reset, 
		    int ack, int frag,
		    const struct pcap_pkthdr *hdr, const u_char* packet)
{
  int i, free=NOT_FOUND, flag, isMalformed;
  lasttime = time;
  flag = (!syn && !reset && !ack && !push && !fin);
  isMalformed = malformed(source,dest);

  /* If we ran out of alarms read new ones */
  if (num_alerts < MIN_ALERTS || (last_alert >=0 && 
				  time > alerts[last_alert].time))
    {
      while(TRUE)
	{
	  char dummy[MAX_ADDRESS];
	  int index = (last_alert + 1) % MAX_ALERTS;
	  fscanf(victimfile_in, "%lf %d %u %s",
		 &alerts[index].time,
		 &alerts[index].attacktype,
		 &alerts[index].victim,
		 dummy);
	  if (feof(victimfile_in))
	    break;
	  last_alert = (last_alert+1) % MAX_ALERTS;
	  num_alerts++;
	  if (num_alerts == MAX_ALERTS || 
	      alerts[last_alert].time > time + STALE_PERIOD)
	    break;
	}
    }

  /* First see if we have an attack trace for this victim,
     at the same time remember first free trace */
  for(i=0;i<MAX_VICTIMS;i++)
    {
      if (attack_trace[i] == 0)
	{
	  if(free == NOT_FOUND)
	    free = i;
	  continue;
	}

      /* A trace must match the type of the traffic - protocol and
	 flags for TCP traffic, malformed indication for malformed
	 traffic, etc. */
      if (isMatch(attack_trace[i], dest, flag, packettype, syn, isMalformed, frag))
	break;

      /* Check here if we can close the trace */
      if (time - attack_trace[i]->last_time > STALE_PERIOD)
	{
	  closeTrace(attack_trace[i]);
	  attack_trace[i]=0;
	  if (free == NOT_FOUND)
	    free = i;
	}
    }

  if (i<MAX_VICTIMS)
    {
      insert_packet(attack_trace[i], hdr, packet, time, packettype);
    }
  else if (free != NOT_FOUND)
    {
      make_new_trace(free, dest, flag, packettype, syn, isMalformed, frag);
      insert_packet(attack_trace[free], hdr, packet, time, packettype);
    }
  else
    {
      printf("Alert, more than %d simultaneous victims, enlarge the limit\n",
	     MAX_VICTIMS);
      exit(1);
    }
}

/* Explains the usage of the program
   input: name of the program
   output: a warning about using the program 
   and command line arguments are printed to the screen */

void usage(char* prg)
{
  printf("Usage %s tcpdump-attack-input-file victim-input-file human-readable-output-file machine-readable-output-file\n", prg);
  exit(1);
}


int main(int argc, char** argv)
{
  int i;
 
  if (argc < 5)
    usage(argv[0]);

  /* Open files for attack alerts */
  victimfile_in = fopen(argv[2],"r");
  human = fopen(argv[3],"w");
  machine = fopen(argv[4],"w");
  if (!victimfile_in || !human || !machine)
    {
      printf("Could not open victim files  %s=%d, %s=%d and %s=%d\n", 
	     argv[2], victimfile_in, argv[3], human, argv[4], machine);
      exit(1);
    }
  /* Initialize buffers for storing information for sort */
  for(i=0;i<MAX_VICTIMS;i++)
    attack_trace[i] = 0;

  /* Initialize a hash for counting sources */
  create_hash(&desttable,DEST_TABLE_SIZE,DEST_TABLE_SIZE,DEST_TABLE_SIZE,1,0,
	      sizeof(struct sdest_key),sizeof(struct sdest_data),(int (*)(void *))sdest_clean,(int (*)(void *,void *))sdest_compare,
	      ( unsigned long(*)(void *, unsigned long ))sdest_hash1,(unsigned long(*)(unsigned long, void*, unsigned long))sdest_hash2);

  read_pcap(1, argv); 
  for(i=0;i<MAX_VICTIMS;i++)
    if (attack_trace[i])
      {
	closeTrace(attack_trace[i]);
      }
  fflush(human);
  fflush(machine);
  fclose(human);
  fclose(machine);
  return 0;  
}











